/**
 * Copyright 2017 Confluent Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#ifndef _POSIX_C_SOURCE
#define _POSIX_C_SOURCE 1 /* sigemptyset(), etc */
#endif
#include <signal.h>


#include "c3metrics.h"
#include "interceptors.h"
#include "pb_encode.h"
#include "monitoring_pb_extension.h"


static void c3ic_send_metrics (c3ic_t *ic);




/**
 * @brief Interceptor logging and debugging function.
 *
 * @param ic Optional interceptor instance
 * @param debug Bool: is debug log
 * @param fmt Log message
 */

void c3ic_log0 (const c3ic_t *ic, int debug, const char *fmt, ...) {
        va_list ap;
        char buf[512];

        if (debug && (!ic || !ic->debug))
                return;

        va_start(ap, fmt);
        ic_vsnprintf(buf, sizeof(buf), fmt, ap);
        va_end(ap);

        fprintf(stderr, "C3IC: %s: %s\n", ic ? ic->clientid : "", buf);
}


/**
 * @brief Interceptor background thread main loop.
 */
static int c3ic_thread_main (void *arg) {
        c3ic_t *ic = arg;

        c3ic_dbg(ic, "Background thread with interval %d, C3 topic %s",
                 ic->publish_ms, rd_kafka_topic_name(ic->rkt));

        mtx_lock(&ic->lock);
        while (!ic->terminate) {
                /* FIXME: the cnd may have a spurious wakeup in which case
                 *        we should adjust the remaining time or use a
                 *        fixed absolute time scheme. */
                if (cnd_timedwait_ms(&ic->cnd, &ic->lock, ic->publish_ms) !=
                    thrd_timedout)
                        continue;

                c3ic_send_metrics(ic);

        }

        /* Send final metrics */
        c3ic_send_metrics(ic);

        mtx_unlock(&ic->lock);

        if (ic->buf)
                free(ic->buf);

        c3ic_dbg(ic, "Background thread terminating");
        return 0;
}




/**
 * @brief Start a new session by generating a new UUID and
 *        resetting the expiry time.
 *
 * @locality ic background thread
 */
static void c3ic_start_session (c3ic_t *ic, int64_t now) {
        ic->session_expiry = now + ic->session_duration_ms;
        ic_uuid_str(ic->session, sizeof(ic->session));
        c3ic_dbg(ic, "Starting new session %s (expires in %"PRId64"ms)",
                 ic->session, ic->session_duration_ms);
}



/**
 * @brief Destroy interceptor instance.
 *
 * @locality application thread calling rd_kafka_destroy()
 */
void c3ic_destroy (c3ic_t *ic) {
        int res;

        c3ic_dbg(ic, "Destroying interceptor");

        /* Signal termination to background thread and wait for it to exit. */
        mtx_lock(&ic->lock);
        ic->terminate = 1;
        cnd_signal(&ic->cnd);
        mtx_unlock(&ic->lock);

        thrd_join(ic->thread, &res);

        c3ic_dbg(ic, "Destroying interceptor: flushing %d message(s)",
                 (int)rd_kafka_outq_len(ic->prod));
        rd_kafka_flush(ic->prod, 1*1000);
        c3ic_dbg(ic, "Destroying interceptor with %d undelivered message(s)",
                 (int)rd_kafka_outq_len(ic->prod));

        tpmap_destroy(&ic->tpm);

        if (ic->clusterid)
                rd_kafka_mem_free(ic->rk, ic->clusterid);
        if (ic->clientid)
                free(ic->clientid);
        if (ic->groupid)
                free(ic->groupid);

        rd_kafka_topic_destroy(ic->rkt);
        rd_kafka_destroy(ic->prod);

        cnd_destroy(&ic->cnd);
        mtx_destroy(&ic->lock);
        free(ic);
}


/**
 * @brief Create new C3 interceptor instance attached to client instance \p rk
 *        using the interceptor config from \p ic_conf
 *
 * @locality application thread calling rd_kafka_new()
 */
c3ic_t *c3ic_new (rd_kafka_t *rk, c3ic_conf_t *ic_conf) {
        c3ic_t *ic;
        char errstr[512];
        rd_kafka_conf_t *conf;
#ifndef _MSC_VER
        sigset_t newset, oldset;
#endif
        char clientid[512];
        char groupid[512];
        size_t sz;

        ic = calloc(1, sizeof(*ic));

        ic->rk = rk;

        /* Add interceptor methods to intercepted client */
        rd_kafka_interceptor_add_on_destroy(rk, IC_NAME, on_destroy, ic);

        /* Take ownership of rdkafka config object which
         * rd_kafka_new() will free. */
        conf = ic_conf->c3_conf;
        ic_conf->c3_conf = NULL;

        /* Since the configuration object is based on the intercepted
         * client's configuration we need to perform some cleanups
         * to make it behave the way the C3 interceptor expects and not to
         * interfere with the intercepted client, namely:
         *  - clear any enabled_events
         *  - clear all callbacks
         *
         * See https://github.com/edenhill/librdkafka/issues/2611
         */
        rd_kafka_conf_set_events(conf, 0);
        rd_kafka_conf_set_dr_cb(conf, NULL);
        rd_kafka_conf_set_dr_msg_cb(conf, NULL);
        rd_kafka_conf_set_stats_cb(conf, NULL);
        rd_kafka_conf_set_background_event_cb(conf, NULL);
        rd_kafka_conf_set_error_cb(conf, NULL);
        rd_kafka_conf_set_throttle_cb(conf, NULL);
        rd_kafka_conf_set_log_cb(conf, NULL);

        switch (rd_kafka_type(rk))
        {
        case RD_KAFKA_CONSUMER:
                rd_kafka_interceptor_add_on_consume(rk, IC_NAME,
                                                    on_consume, ic);
                ic->clienttype = monitoring_ClientType_CONSUMER;
                sz = sizeof(groupid)-1;
                if (rd_kafka_conf_get(conf, "group.id", groupid, &sz) == RD_KAFKA_CONF_OK) {
                        ic->groupid = ic_strdup(groupid);
                }
                break;
        case RD_KAFKA_PRODUCER:
                rd_kafka_interceptor_add_on_acknowledgement(rk, IC_NAME,
                                                            on_acknowledgement,
                                                            ic);
                /* FALLTHRU */
        default: /* For future compatibility */
                ic->clienttype = monitoring_ClientType_PRODUCER;
        break;

        }

        /* Extract intercepted client's client.id, which is used in metrics,
         * and construct the interceptor client's client.id from it,
         * unless it has been configured explicitly. */
        clientid[sizeof(clientid)-1] = '\0';
        sz = sizeof(clientid)-1;
        if (rd_kafka_conf_get(conf, "client.id", clientid, &sz) !=
            RD_KAFKA_CONF_OK)
                snprintf(clientid, sizeof(clientid), "rdkafka"); /* Default */

        ic->clientid = ic_strdup(clientid);

        if (!*ic_conf->clientid) {
                /* The interceptor client.id has not been explicitly set,
                 * construct it from intercepted client's client.id
                 * by prefixing the interceptor prefix. */
                snprintf(ic_conf->clientid, sizeof(ic_conf->clientid),
                         "%s%s", MONITORING_INTERCEPTOR_PREFIX, ic->clientid);
        }

        rd_kafka_conf_set(conf, "client.id", ic_conf->clientid, NULL, 0);

        ic->debug = ic_conf->debug;
        ic->publish_ms = ic_conf->publish_ms;
        ic->session_duration_ms = ic_conf->session_duration_ms;

        mtx_init(&ic->lock, mtx_plain);
        cnd_init(&ic->cnd);

        c3ic_dbg(ic, "Creating c3 metrics interceptor");

        c3ic_start_session(ic, get_wallclock_timestamp());

        ic->prod = rd_kafka_new(RD_KAFKA_PRODUCER, conf,
                                  errstr, sizeof(errstr));
        if (!ic->prod) {
                c3ic_log(ic, "Failed to create producer: %s", errstr);
                free(ic);
                return NULL;
        }

        ic->rkt = rd_kafka_topic_new(ic->prod, ic_conf->topic, NULL);
        if (!ic->rkt) {
                c3ic_log(ic, "Failed to create topic \"%s\" handle: %s",
                         ic_conf->topic, errstr);
                rd_kafka_destroy(ic->prod);
                cnd_destroy(&ic->cnd);
                mtx_destroy(&ic->lock);
                free(ic);
                return NULL;
        }


        tpmap_init(&ic->tpm, 31);

#ifndef _MSC_VER
        /* Block all signals in newly created thread.
         * To avoid race condition we block all signals in the calling
         * thread, which the new thread will inherit its sigmask from,
         * and then restore the original sigmask of the calling thread when
         * we're done creating the thread.
         * NOTE: term_sig remains unblocked since we use it on termination
         *       to quickly interrupt system calls. */
        sigemptyset(&oldset);
        sigfillset(&newset);
        pthread_sigmask(SIG_SETMASK, &newset, &oldset);
#endif

        /* Create background thread for emitting metrics
         * every interval. */
        if (thrd_create(&ic->thread, c3ic_thread_main, ic) != thrd_success) {
                c3ic_log(ic, "Failed to create thread");
                rd_kafka_topic_destroy(ic->rkt);
                rd_kafka_destroy(ic->prod);
                cnd_destroy(&ic->cnd);
                mtx_destroy(&ic->lock);
                free(ic);
                ic = NULL;
                /* FALLTHRU */
        }

#ifndef _MSC_VER
        /* Restore sigmask of caller */
        pthread_sigmask(SIG_SETMASK, &oldset, NULL);
#endif


        return ic;
}


/**
 * @brief Fill in min and max window fields in monitoring message with
 *        min and max windows recorded in the history.
 *        Min and max windows are omitted if no windows were recorded
 *        in the last 'size' of sequence numbers (sending empty metrics).
 *
 * @remark Also adds window from monitoring message to the history.
 */
static void
c3ic_record_fill_minmax_windows (timebuckets_t *tbs,
                                 monitoring_MonitoringMessage *msg) {
        int i = tbs->tbs_h_i;
        do {
                int64_t window = tbs->tbs_history[i];

#define default_if_unknown(unk,def)                                     \
                ((unk) == TIMEBUCKETS_UNKNOWN_WINDOW ? (def) : (unk))

                if (window != TIMEBUCKETS_UNKNOWN_WINDOW) {
                        if (msg->minWindow == TIMEBUCKETS_UNKNOWN_WINDOW ||
                            window < msg->minWindow) {
                                msg->minWindow = window;
                        }

                        if (msg->maxWindow == TIMEBUCKETS_UNKNOWN_WINDOW ||
                            window > msg->maxWindow) {
                                msg->maxWindow = window;
                        }

                }

                i = (i + 1) % tbs->tbs_h_sz;
        } while (i != tbs->tbs_h_i);

        /* we want to record null windows, so that we always use history of
         * last 'size sequence numbers, independent of how many of them are
         * heartbeat messages without a window */
        timebuckets_history_add(tbs, msg->window);

}


/**
 * @brief Send metrics record message to C3 for the given partition.
 *
 * @locality ic background thread
 */
static void c3ic_send_record (c3ic_t *ic, tp_t *tp,
                              monitoring_MonitoringMessage *msg,
                              int mark_shutdown, int64_t now) {
        size_t size, enc_size;
        pb_ostream_t stream;

        msg->clientType         = ic->clienttype;
        string_encode_callback(&msg->clientId, ic->clientid);
        string_encode_callback(&msg->group, ic->groupid);
        string_encode_callback(&msg->session, ic->session);
        string_encode_callback(&msg->clusterId, ic->clusterid);
        string_encode_callback(&msg->clusterName, NULL);
        msg->sequence           = tp->tbs.tbs_sequence++;
        msg->timestamp          = now;
        string_encode_callback(&msg->topic, tp->topic);
        msg->partition          = tp->partition;
        msg->samplePeriod       = ic->publish_ms;
        if (mark_shutdown) {
                msg->shutdown     = mark_shutdown;
        }
        c3ic_record_fill_minmax_windows(&tp->tbs, msg);

        pb_get_encoded_size(&enc_size, monitoring_MonitoringMessage_fields, msg);
        size = sizeof(magic_byte_t) + enc_size;
        if (size > ic->buf_size) {
                /* Increase shared buffer if needed */
                if (ic->buf)
                        free(ic->buf);
                ic->buf_size = (size + 16) * 2;
                ic->buf = malloc(ic->buf_size);
                if (!ic->buf) {
                        c3ic_log(ic, "Failed to allocate %zd bytes "
                                 "for protobuf: %s",
                                 ic->buf_size, strerror(errno));
                        ic->buf_size = 0;
                        return;
                }
        }

        ic->buf[0] = MAGIC_BYTE_PROTOBUF;

        stream = pb_ostream_from_buffer(ic->buf+sizeof(magic_byte_t), enc_size);

        pb_encode(&stream, monitoring_MonitoringMessage_fields, msg);

        c3ic_dbg(ic, "Producing %srecord (%zd bytes) to "
                 "C3 topic %s for %s [%"PRId32"] window %"PRId64,
                 msg->count ? "" : "empty ",
                 size, rd_kafka_topic_name(ic->rkt), tp->topic, tp->partition,
                 msg->window);

        /* Produce record to C3 cluster */
        if (rd_kafka_produce(ic->rkt, RD_KAFKA_PARTITION_UA,
                             RD_KAFKA_MSG_F_COPY,
                             ic->buf, size, NULL, 0, NULL) == -1) {
                /* Log but otherwise ignore send errors */
                c3ic_dbg(ic, "C3 produce() failed: %s",
                         rd_kafka_err2str(rd_kafka_last_error()));
        }
}


/**
 * @brief Send empty heartbeat metrics record message to C3
 */
static void c3ic_send_empty_record (c3ic_t *ic, tp_t *tp, int mark_shutdown,
                                    int64_t now) {
        monitoring_MonitoringMessage msg = monitoring_MonitoringMessage_init_default;
        msg.type = monitoring_MessageType_HEARTBEAT;
        c3ic_send_record(ic, tp, &msg, mark_shutdown, now);
}




/**
 * @brief Send metrics to C3 for all known topic+partitions.
 *
 * @locks ic->lock MUST be held
 * @locality ic background thread
 */
static void c3ic_send_metrics (c3ic_t *ic) {
        tp_t *tp;
        int64_t now = get_wallclock_timestamp();
        int mark_shutdown = 0;
        int session_expired = 0;

        /* Get the Clusterid if not already cached. */
        if (!ic->clusterid) {
                ic->clusterid = rd_kafka_clusterid(ic->rk, 3*1000);
                if (!ic->clusterid) {
                        c3ic_log(ic, "Unable to acquire ClusterId from "
                                 "intercepted client "
                                 "(missing api.version.request=true?)");
                } else {
                        c3ic_dbg(ic, "Acquired ClusterId %s", ic->clusterid);
                }
        }

        /* Check if the session expired, if so mark the last window
         * for any topic+partition with shutdown=true */
        if (ic->session_expiry <= now)
                session_expired = 1;

        if (ic->terminate || session_expired)
                mark_shutdown = 1;

        TAILQ_FOREACH(tp, &ic->tpm.tpm_flat, flat_link) {
                timebucket_t *tb, *next;

                if (TAILQ_EMPTY(&tp->tbs.tbs_flat)) {
                        /* Send empty heartbeat metrics message */
                        c3ic_send_empty_record(ic, tp, mark_shutdown, now);

                } else {
                        TAILQ_FOREACH_SAFE(tb, &tp->tbs.tbs_flat,
                                           tb_flat_link, next) {
                                int is_last = !next;

                                c3ic_send_record(ic, tp, &tb->tb_m,
                                                 is_last && mark_shutdown, now);

                                timebucket_destroy(tb);
                        }

                        /* Clear and reset the bucket list. */
                        timebuckets_clear(&tp->tbs,
                                          0/* dont destroy buckets
                                            * (destroyed above) */);
                }

                if (session_expired)
                        tp->tbs.tbs_sequence = 0;
        }

        if (session_expired)
                c3ic_start_session(ic, now);
}



/**
 * @brief Record message metric into monitoring metrics
 *        for the given topic-partition.
 *
 * @param tp        Topic+partition
 * @param timestamp Message timestamp
 * @param size      Message size (key plus value, pre-compression)
 * @param latency   Latency of the message if available,
 *                  otherwise non-positive value
 *
 * @locks ic->lock MUST be held
 * @locality any thread
  *
 */
static void c3ic_tp_record (c3ic_t *ic, tp_t *tp,
                            int64_t timestamp, size_t size, int64_t latency) {
        timebucket_t *tb;

        tb = timebucket_get(&tp->tbs, timestamp);

        tb->tb_m.count++;
        tb->tb_m.aggregateBytes += (int64_t)size;
        if (latency < tb->tb_m.minLatency) {
                tb->tb_m.minLatency = latency;
        }

        if (latency >= tb->tb_m.maxLatency) {
                tb->tb_m.maxLatency = latency;
        }


        tb->tb_m.totalLatency += latency;
}



/**
 * @brief Record message metric from producer or consumer message interceptor
 *
 * @locality application thread
 */
void c3ic_record_message_metric (c3ic_t *ic,
                                 const rd_kafka_message_t *rkmessage) {
        size_t size = rkmessage->len + rkmessage->key_len;
        tp_t *tp;
        int64_t timestamp, latency = 0;
        rd_kafka_timestamp_type_t tstype;

        timestamp = rd_kafka_message_timestamp(rkmessage, &tstype);

        /* Ignore messages without a valid timestamp since they can't be mapped
         * to a proper bucket and thus has no meaningful statistical value.
         * This only happens when interacting with <0.10 brokers or
         * api.version.request has been disabled or the ApiVersionRequest
         * has failed. */
        if (tstype == RD_KAFKA_TIMESTAMP_NOT_AVAILABLE || timestamp <= 0) {
                int64_t now = get_wallclock_timestamp();
                mtx_lock(&ic->lock);
                ic->log_msg_ignore_cnt++;
                if (ic->ts_log_msg_ignore + (1000 * 60 * 60) < now) {
                        c3ic_log(ic,
                                 "Ignoring message on topic \"%s\" [%"PRId32"]"
                                 " at offset %"PRId64" without valid timestamp "
                                 "(%"PRId64" messages ignored)",
                                 rd_kafka_topic_name(rkmessage->rkt),
                                 rkmessage->partition, rkmessage->offset,
                                 ic->log_msg_ignore_cnt);
                        ic->ts_log_msg_ignore = now;
                }
                mtx_unlock(&ic->lock);
                return;
        }

        if (ic->clienttype == monitoring_ClientType_CONSUMER) {
                /* Consumer.
                 * Latency calculation is end-to-end based on the producer
                 * message timestamp. */

                latency = get_wallclock_timestamp() - timestamp;

        } else {
                /* Producer.
                 * Latency calculation is the full produce cycle latency. */

                latency = rd_kafka_message_latency(rkmessage) / 1000;
        }

#if 0 /* noisy */
        c3ic_dbg(ic, "on_consume() message timestamp %"PRId64" (%d), "
                 "latency %"PRId64,
                 timestamp, tstype, latency);
#endif

        if (latency < 0)
                latency = 0;

        mtx_lock(&ic->lock);

        tp = tp_get(&ic->tpm, rd_kafka_topic_name(rkmessage->rkt),
                    rkmessage->partition, ic->publish_ms);

        c3ic_tp_record(ic, tp, timestamp, size, latency);

        mtx_unlock(&ic->lock);
}




