/**
 * Copyright 2017 Confluent Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#pragma once

#include <librdkafka/rdkafka.h>

#include "platform.h"
#include "tinycthread.h"
#include "tpmap.h"


/* Configuration prefix */
#define MONITORING_INTERCEPTOR_PREFIX "confluent.monitoring.interceptor."

/* Interceptor name */
#define IC_NAME "c3metrics"

/* Default publish / metric window period in milliseconds */
#define DEFAULT_PUBLISH_PERIOD (15*1000)

/* Default C3 topic */
#define DEFAULT_TOPIC "_confluent-monitoring"

#define DEFAULT_SESSION_DURATION (10*60*1000)  /* 10 minutes */

#define MAGIC_BYTE_PROTOBUF 127

typedef char magic_byte_t;


typedef struct c3ic_overrides_conf_s {
        char **prop_names;
        int capacity;
        int size;
} c3ic_overrides_conf_t;

typedef struct c3ic_conf_s {
        rd_kafka_conf_t *c3_conf;       /* Configuration for C3 producer handle */
        c3ic_overrides_conf_t overrides; /* Tracks client configuration overrides */
        char *topic;                    /* C3 topic to produce to */
        int publish_ms;                 /* C3 publishing period / window size */
        int session_duration_ms;        /* C3 session duration */
        int debug;                      /* Interceptor debugging */
        /* C3 producer's client.id rules:
         *    a) if explicitly configured, leave as is.
         *    b) else use intercepted client's client.id (possibly default)
         *      prefixed with MONITORING_INTERCEPTOR_PREFIX. */
        char  clientid[512];      /* intercepted client.id prefixed with
                                   * MONITORING_INTERCEPTOR_PREFIX */
} c3ic_conf_t;



/**
 * @brief C3 metrics interceptor
 */
typedef struct c3ic_s {
        /*
         * Any (e.g., application, interceptor, librdkafka) thread(s)
         * Locking needed.
         */
        thrd_t thread;          /* Background thread */
        cnd_t  cnd;             /* Application -> background thread signaling */
        mtx_t  lock;            /* Global lock for this instance */
        int    terminate;       /* Signals termination to background thread */
        int64_t ts_log_msg_ignore;  /* Timestamp of last ignored message log */
        int64_t log_msg_ignore_cnt; /* Number of ignored messages */

        tpmap_t tpm;            /* Topic+Partition -> Time bucket map */

        /*
         * Local to background thread (after initialization).
         * No need for locking.
         */
        int    publish_ms;      /* Metric window size / period */
        int    session_duration_ms; /* C3 session duration */
        int    debug;           /* Debugging enabled */


        rd_kafka_t *rk;         /* Intercepted client handle */
        rd_kafka_t *prod;       /* C3 metrics producer */
        rd_kafka_topic_t *rkt;  /* C3 topic to produce to */

        monitoring_ClientType clienttype;
        char *clientid;
        char *groupid;
        char  session[32+1];    /* UUID */
        int64_t session_expiry; /* Absolute time the session expires. */
        char *clusterid;

        uint8_t *buf;           /* Shared protobuf serializer buffer */
        size_t buf_size;        /* .. buffer size */

} c3ic_t;


void c3ic_log0 (const c3ic_t *ic, int debug, const char *fmt, ...);

#define c3ic_log(IC,...) c3ic_log0(IC, 0, __VA_ARGS__)
#define c3ic_dbg(IC,...) c3ic_log0(IC, 1, __VA_ARGS__)

void c3ic_record_message_metric (c3ic_t *ic,
                                 const rd_kafka_message_t *rkmessage);

c3ic_t *c3ic_new (rd_kafka_t *rk, c3ic_conf_t *ic_conf);
void c3ic_destroy (c3ic_t *ic);
