/**
 * Copyright 2017 Confluent Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <assert.h>

#include "c3metrics.h"
#include "interceptors.h"


static void conf_init0 (rd_kafka_conf_t *conf);

static void c3ic_overrides_set(c3ic_overrides_conf_t *ic_ov, const char *prop_name);
static const char *c3ic_normalize_prop_name(const char *prop);
static int c3ic_overrides_is_set(c3ic_overrides_conf_t *ic_ov, const char *match);

/**
 * @brief Producer: This method is called when the record sent to the server has
 *        been acknowledged, or when sending the record fails before it
 *        gets sent to the server.
 *
 * Currently, this implementation records only successful acks.
 *
 * @param message The metadata for the message that was sent (i.e.
 *                the partition and offset). If
 *                error occurred, metadata is only partially filled
 *                (topic, and maybe partition).
 *
 * @locality application thread or internal librdkafka thread (if no dr_msg_cb)
 */

rd_kafka_resp_err_t on_acknowledgement (rd_kafka_t *rk,
                                        rd_kafka_message_t *rkmessage,
                                        void *ic_opaque) {
        c3ic_t *ic = ic_opaque;

        /* Ignore errored transmissions for now */
        if (rkmessage->err)
                return RD_KAFKA_RESP_ERR_NO_ERROR;

        c3ic_record_message_metric(ic, rkmessage);

        return RD_KAFKA_RESP_ERR_NO_ERROR;

}


/**
 * @brief Consumer: called for each message just prior to passing it to the
 *        the application.
 *
 * @locality application thread
 */
rd_kafka_resp_err_t on_consume (rd_kafka_t *rk,
                                rd_kafka_message_t *rkmessage,
                                void *ic_opaque) {
        c3ic_t *ic = ic_opaque;

        /* This callback is currently not called for consumer errors in
         * librdkafka v0.9.5, but we handle it anyway to be future proof. */
        if (rkmessage->err)
                return RD_KAFKA_RESP_ERR_NO_ERROR;

        c3ic_record_message_metric(ic, rkmessage);

        return RD_KAFKA_RESP_ERR_NO_ERROR;
}


/**
 * @brief Client handle on_destroy() interceptor.
 */
rd_kafka_resp_err_t on_destroy (rd_kafka_t *rk, void *ic_opaque) {
        c3ic_t *ic = ic_opaque;

        c3ic_destroy(ic);

        return RD_KAFKA_RESP_ERR_NO_ERROR;
}


/**
 * @brief Client handle on_new() interceptor.
 */
rd_kafka_resp_err_t on_new (rd_kafka_t *rk, const rd_kafka_conf_t *conf,
                            void *ic_opaque, char *errstr, size_t errstr_size) {
        c3ic_new(rk, (c3ic_conf_t *)ic_opaque);

        return RD_KAFKA_RESP_ERR_NO_ERROR;
}


/**
 * @brief Parses a supposed integer string,
 *        returning the parsed value in \p *ret.
 * @returns -1 on parse error (not an integer), or 0 on success.
 */
static int parse_int_val (const char *str, int *ret, int vmin, int vmax) {
        const char *end;
        int iv = strtol(str, (char **)&end, 0);

        if (end == str || iv < vmin || iv > vmax)
                return -1;

        *ret = iv;

        return 0;
}


/**
 * @brief on_conf_set interceptor for setting interceptor-specific
 *        configuration properties.
 *
 * Any configuration property prefixed with MONITORING_INTERCEPTOR_PREFIX
 * will be set on the C3 producer's configuration object, apart from
 * a small set of interceptor-specific properties that are set on
 * the c3ic_conf_t object.
 */
rd_kafka_conf_res_t on_conf_set (rd_kafka_conf_t *conf,
                                 const char *name, const char *val,
                                 char *errstr, size_t errstr_size,
                                 void *ic_opaque) {
        c3ic_conf_t *ic_conf = ic_opaque;
        const char *subname;
        size_t pfxlen = strlen(MONITORING_INTERCEPTOR_PREFIX);

        if (strncmp(name, MONITORING_INTERCEPTOR_PREFIX, pfxlen)) {
                /* Apply intercepted client's configuration
                 * to the interceptor configuration as well;
                 * Skips property if it was previously set
                 * with the interceptor prefix.
                 * Ignore errors. */
                if (!c3ic_overrides_is_set(&ic_conf->overrides, name))
                        rd_kafka_conf_set(ic_conf->c3_conf, name, val, NULL, 0);

                return RD_KAFKA_CONF_UNKNOWN;
        }

        subname = name + (int)pfxlen;

        /* Interceptor-specific configuration properties. */
        if (!ic_strcasecmp(subname, "publishms")) {
                if (!val)
                        ic_conf->publish_ms = DEFAULT_PUBLISH_PERIOD;
                else if (parse_int_val(val, &ic_conf->publish_ms,
                                       1, 86400*1000) == -1)
                        goto inval;

        } else if (!ic_strcasecmp(subname, "sessiondurationms")) {
                if (!val)
                        ic_conf->session_duration_ms = DEFAULT_SESSION_DURATION;
                else if (parse_int_val(val, &ic_conf->session_duration_ms,
                                       1, INT_MAX) == -1)
                        goto inval;

        } else if (!strcmp(subname, "topic")) {
                if (ic_conf->topic) {
                        free(ic_conf->topic);
                        ic_conf->topic = NULL;
                }
                if (val) {
                        if (!*val)
                                goto inval;
                        ic_conf->topic = ic_strdup(val);
                } else {
                        ic_conf->topic = ic_strdup(DEFAULT_TOPIC);
                }

        } else if (!strcmp(subname, "icdebug")) {
                if (!val || !ic_strcasecmp(val, "false"))
                        ic_conf->debug = 0;
                else
                        ic_conf->debug = 1;

        } else if (!strcmp(subname, "client.id")) {
                /* Special handling for client.id */
                if (val) {
                        strncpy(ic_conf->clientid, val,
                                sizeof(ic_conf->clientid)-1);
                } else {
                        /* client.id was reset */
                        ic_conf->clientid[0] = '\0';
                }

        } else {
                c3ic_overrides_set(&ic_conf->overrides, subname);
                /* Set standard librdkafka properties on C3 producer's config */
                return rd_kafka_conf_set(ic_conf->c3_conf,
                                         subname, val, errstr, errstr_size);
        }

        return RD_KAFKA_CONF_OK;

 inval:
        snprintf(errstr, errstr_size,
                 "Invalid value for configuration property \"%s\"", name);
        return RD_KAFKA_CONF_INVALID;
}


/**
 * @brief Copy configuration. The actual properties will be copied
 *        by librdkafka and applied through conf_set() so we only
 *        need to set up our own ic_conf object for the new_conf here.
 */
rd_kafka_resp_err_t
on_conf_dup (rd_kafka_conf_t *new_conf, const rd_kafka_conf_t *old_conf,
             size_t filter_cnt, const char **filter, void *ic_opaque) {
        conf_init0(new_conf);
        return RD_KAFKA_RESP_ERR_NO_ERROR;
}


/**
 * @brief Clean up any overrides which may have been set.
 */
static void c3ic_overrides_destroy(c3ic_overrides_conf_t *ic_ov) {
        int idx;

        for (idx = 0; idx < ic_ov->size; idx++)
                free((char *)ic_ov->prop_names[idx]);
        free(ic_ov->prop_names);
}


/**
 * @brief on_conf_destroy() interceptor. Free our c3ic_conf_t object.
 */
rd_kafka_resp_err_t
on_conf_destroy (void *ic_opaque) {
        c3ic_conf_t *ic_conf = ic_opaque;

        if (ic_conf->c3_conf)
                rd_kafka_conf_destroy(ic_conf->c3_conf);
        if (ic_conf->topic)
                free(ic_conf->topic);
        if (ic_conf->overrides.prop_names)
                c3ic_overrides_destroy(&ic_conf->overrides);
        free(ic_conf);

        return RD_KAFKA_RESP_ERR_NO_ERROR;
}


/**
 * @brief Set up new ic_conf object, adding default config for interceptor
 *        and setting up config interceptors.
 */
static void conf_init0 (rd_kafka_conf_t *conf) {
        c3ic_conf_t *ic_conf;
        rd_kafka_conf_t *c3_conf;
        const char *conf_dup_filter[] = {
                "plugin.library.paths",
                "confluent.monitoring.interceptor"
        };

        /* Create new interceptor configuration handle */
        ic_conf = calloc(1, sizeof(*ic_conf));

        /* Set defaults */
        ic_conf->publish_ms = DEFAULT_PUBLISH_PERIOD;
        ic_conf->session_duration_ms = DEFAULT_SESSION_DURATION;
        ic_conf->topic = ic_strdup(DEFAULT_TOPIC);

        /* Set C3 producer default config based on the intercepted
         * client's configuration, filter out any properties that might
         * get in the way, such as plugin libraries and interceptor config. */
        c3_conf = rd_kafka_conf_dup_filter(conf,
                                           sizeof(conf_dup_filter)/
                                           sizeof(*conf_dup_filter),
                                           conf_dup_filter);

        ic_conf->c3_conf = c3_conf;
        rd_kafka_conf_set(c3_conf, "api.version.request", "true", NULL, 0);
        rd_kafka_conf_set(c3_conf, "log_level", "3", NULL, 0);
        rd_kafka_conf_set(c3_conf, "log.connection.close", "false", NULL, 0);
        rd_kafka_conf_set(c3_conf, "queue.buffering.max.ms", "100", NULL, 0);

        /* Set up config and new() interceptors for intercepted client. */
        rd_kafka_conf_interceptor_add_on_conf_set(conf, IC_NAME, on_conf_set,
                                                  ic_conf);
        rd_kafka_conf_interceptor_add_on_conf_dup(conf, IC_NAME, on_conf_dup,
                                                  ic_conf);
        rd_kafka_conf_interceptor_add_on_conf_destroy(conf, IC_NAME,
                                                      on_conf_destroy, ic_conf);
        rd_kafka_conf_interceptor_add_on_new(conf, IC_NAME, on_new, ic_conf);
}



/**
 * @brief Plugin entry point, called when library is referenced
 *        through `plugin.library.paths` client configuration.
 *
 * @remark Might be called multiple times if configuration object is copied.
 */
DLL_EXPORT
rd_kafka_resp_err_t conf_init (rd_kafka_conf_t *conf,
                               void **plug_opaquep,
                               char *errstr, size_t errstr_size) {
        conf_init0(conf);
        return RD_KAFKA_RESP_ERR_NO_ERROR;
}



/**
 * @brief Sets the provided property in the configuration overrides.
 */
static void c3ic_overrides_set(c3ic_overrides_conf_t *ic_ov, const char *prop_name) {
        if (ic_ov->capacity == ic_ov->size) {
                ic_ov->capacity += ic_ov->capacity + 1;
                ic_ov->prop_names = realloc(ic_ov->prop_names,
                                            sizeof(*ic_ov->prop_names)
                                            * (ic_ov->capacity));
                assert(ic_ov->prop_names);
        }
        ic_ov->prop_names[ic_ov->size++] =
                ic_strdup(c3ic_normalize_prop_name(prop_name));
}

/**
 * @brief Checks if interceptor override already set.
 * @returns true(non-zero) if found else false(zero)
 */
static int c3ic_overrides_is_set(c3ic_overrides_conf_t *ic_ov,
                                 const char *match) {
        const char *normal_match = c3ic_normalize_prop_name(match);
        int idx;

        for (idx = 0; idx < ic_ov->size; idx++)
                if (!strcmp(normal_match, ic_ov->prop_names[idx]))
                        return 1;
        return 0;
}

/**
 * Collection of rdkafka_conf alias, standard name tuples.
 */
static const char *c3ic_conf_synonyms[] = {
        "bootstrap.servers", "metadata.broker.list",
        "max.in.flight", "max.in.flight.requests.per.connection",
        "sasl.mechanism", "sasl.mechanisms",
        "max.partition.fetch.bytes", "fetch.message.max.bytes",
        "linger.ms", "queue.buffering.max.ms",
        "retries", "message.send.max.retries",
        "acks", "request.required.acks",
        "delivery.timeout.ms", "message.timeout.ms",
        "compression.type", "compression.codec",
        "enable.auto.commit", "auto.commit.enable",
        NULL,
};

/**
 * @brief Converts rdkafka_conf alias to its standard name
 */
static const char *c3ic_normalize_prop_name(const char *match) {
        const char **prop = c3ic_conf_synonyms;

        for (; *prop; prop += 2) {
                if (!strcmp(*prop, match)) {
                        const char *synonym = *(prop+1);

                        return synonym;
                }
        }
        return match;
}
