/**
 * Copyright 2017 Confluent Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "platform.h"

#include <stdio.h>
#include <assert.h>




#ifdef __APPLE__
#include <CoreFoundation/CFUUID.h>  /* For CFUUID */

#elif _MSC_VER

#else
/* Linux */
#include <uuid/uuid.h>
#endif


#ifdef _MSC_VER
/**
 * @brief gettimeofday() for win32
 */
int ic_gettimeofday (struct timeval *tv, struct timezone *tz) {
        SYSTEMTIME st;
        FILETIME   ft;
        ULARGE_INTEGER d;

        SystemTimeToFileTime(&st, &ft);
        d.HighPart = ft.dwHighDateTime;
        d.LowPart  = ft.dwLowDateTime;
        tv->tv_sec  = (long)((d.QuadPart - 116444736000000000llu) / 10000000L);
        tv->tv_usec = (long)(st.wMilliseconds * 1000);

        return 0;
}
#endif



int ic_vsnprintf (char *str, size_t size, const char *format, va_list ap) {
        int cnt = -1;

#ifdef _MSC_VER
        cnt = _vsnprintf_s(str, size, _TRUNCATE, format, ap);
#else
        cnt = vsnprintf(str, size, format, ap);
#endif

        return cnt;
}



/**
 * @brief Generate a UUID2 (random) string.
 *
 * @param dst_size MUST be at least 16*2+1 bytes
 */
void ic_uuid_str (char *dst, size_t dst_size) {
        static const char *hex = "0123456789abcdef";
        const char *s;
        int i;

#ifdef __APPLE__
        /*
         * OSX: Use CFUUID
         */
        CFUUIDRef uuid = CFUUIDCreate(NULL);
        CFUUIDBytes ub = CFUUIDGetUUIDBytes(uuid);
        CFRelease(uuid);
        s = (const char *)&ub;

#elif _MSC_VER
        /*
         * Win32: Use Guid
         */
        GUID uuid;
        CoCreateGuid(&uuid);
        s = (const char *)&uuid;
#else
        /*
         * Linux: use libuuid
         */
        uuid_t uuid;
        uuid_generate_random(uuid);
        s = (const char *)uuid;
#endif

        assert(dst_size >= 16*2+1);

        for (i = 0 ; i < 16 ; i++) {
                dst[i*2]   = hex[(s[i]>>4) & 0xF];
                dst[i*2+1] = hex[s[i] & 0xF];
        }
        dst[i*2] = '\0';
}
