/**
 * Copyright 2017 Confluent Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * Time window buckets
 *
 * Consists of a map indexed by base timestamp,
 * each bucket's window corresponds to timestamps in the range:
 *    base_timestamp..+sample_period
 *
 */


#include <stdlib.h>
#include <string.h>

#include "timebucket.h"
#include "monitoring_pb_extension.h"



/**
 * @returns The timestamp base of a timestamp
 */
static __inline int64_t timebuckets_timestamp_base (const timebuckets_t *tbs,
                                                    int64_t timestamp) {
        return (timestamp - (timestamp % tbs->tbs_win_size));
}

/**
 * @brief Hash key function for mapping a timestamp_base to a hash index.
 */
static __inline int timebuckets_hash (const timebuckets_t *tbs,
                                      int64_t timestamp_base) {
        return (timestamp_base / tbs->tbs_win_size) % tbs->tbs_size;
}


/**
 * @brief Free and destroy timebucket.
 * @remark Will NOT unlink from tbs
 */
void timebucket_destroy (timebucket_t *tb) {
        free(tb);
}


/**
 * @brief Initialize history array to default (unknown window) values.
 */
static void timebuckets_history_init (timebuckets_t *tbs) {
        int i;
        tbs->tbs_h_i = 0;
        for (i = 0 ; i < tbs->tbs_h_sz ; i++)
                tbs->tbs_history[i] = TIMEBUCKETS_UNKNOWN_WINDOW;
}


void timebuckets_init (timebuckets_t *tbs, int size, int64_t win_size) {
        int i;

        memset(tbs, 0, sizeof(*tbs));
        tbs->tbs_size = size;
        tbs->tbs_win_size = win_size;
        tbs->tbs_h_sz = TIMEBUCKETS_HISTORY_SIZE;

        timebuckets_history_init(tbs);

        tbs->tbs_buckets = malloc(tbs->tbs_size * sizeof(*tbs->tbs_buckets));
        for (i = 0 ; i < tbs->tbs_size ; i++)
                TAILQ_INIT(&tbs->tbs_buckets[i]);

        TAILQ_INIT(&tbs->tbs_flat);

}


void timebuckets_clear (timebuckets_t *tbs, int destroy_buckets) {
        int i;
        for (i = 0 ; i < tbs->tbs_size ; i++) {
                timebucket_t *tb, *tmp;

                if (destroy_buckets) {
                        TAILQ_FOREACH_SAFE(tb, &tbs->tbs_buckets[i],
                                           tb_link, tmp) {
                                timebucket_destroy(tb);
                        }
                }

                TAILQ_INIT(&tbs->tbs_buckets[i]);
        }

        TAILQ_INIT(&tbs->tbs_flat);

        timebuckets_history_init(tbs);
}


void timebuckets_destroy (timebuckets_t *tbs) {
        timebuckets_clear(tbs, 1/*destroy_buckets*/);
        free(tbs->tbs_buckets);
}


static __inline int timebucket_cmp (const void *_a, const void *_b) {
        const timebucket_t *a = _a, *b = _b;

        return (int)(a->tb_base - b->tb_base);
}


/**
 * @brief Get or add timebucket
 *
 * Buckets are inserted in sorted (descending) order, which hopefully improves
 * lookup times.
 */
timebucket_t *timebucket_get (timebuckets_t *tbs, int64_t timestamp) {
        int64_t timestamp_base = timebuckets_timestamp_base(tbs, timestamp);
        int bkt = timebuckets_hash(tbs, timestamp_base);
        timebucket_t *elem;
        timebucket_t *tb;
        timebucket_t skel;

        skel.tb_base = timestamp_base;

        TAILQ_FOREACH(elem, &tbs->tbs_buckets[bkt], tb_link) {
                int r = timebucket_cmp(&skel, elem);

                if (r == 0) /* Window found, return. */
                        return elem;
                else if (r > 0)
                        break; /* Key not found, break out and add it
                                * at this position. */
        }

        /* Create new tb and add it to its sorted location in the bucket. */
        tb = calloc(1, sizeof(*tb));
        tb->tb_base = timestamp_base;

        monitoring_message_init(&tb->tb_m);

        tb->tb_m.type               = monitoring_MessageType_NORMAL;
        tb->tb_m.window             = tb->tb_base;

        if (elem)
                TAILQ_INSERT_BEFORE(elem, tb, tb_link);
        else
                TAILQ_INSERT_TAIL(&tbs->tbs_buckets[bkt], tb, tb_link);

        TAILQ_INSERT_TAIL(&tbs->tbs_flat, tb, tb_flat_link);

        return tb;
}




