/**
 * Copyright 2017 Confluent Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/**
 * @brief Topic+Partition map
 *
 * Data value is currently hard-code to timebuckets_t
 */

#include <stdlib.h>
#include <string.h>

#include "tpmap.h"


/**
 * @brief Destroy topic-partition and its timebuckets.
 * @remark Will NOT unlink the tp.
 */
void tp_destroy (tp_t *tp) {
        timebuckets_destroy(&tp->tbs);
        free(tp->topic);
        free(tp);
}


static __inline int tp_cmp (const void *_a, const void *_b) {
        const tp_t *a = _a, *b = _b;
        int r;

        r = (int)(a->partition - b->partition);
        if (!r)
                r = strcmp(a->topic, b->topic);

        return r;
}

/**
 * @brief tp hasher, uses djb hashing on partition + topic
 */
static __inline int tp_hash (const tpmap_t *tpm, const tp_t *tp) {
        unsigned int hash = 5381;
        const char *t = tp->topic;

        hash = ((hash << 5) + hash) + (unsigned int)tp->partition;

        for (t = tp->topic ; *t ; t++)
                hash = ((hash << 5) + hash) + (unsigned int)*t;

        return hash % tpm->tpm_size;
}



/**
 * @brief Get or add topic+partition to tpmap.
 */
tp_t *tp_get (tpmap_t *tpm, const char *topic, int32_t partition,
              int win_size) {
        tp_t *elem, *tp;
        tp_t skel;
        int i;

        skel.partition = partition;
        skel.topic = (char *)topic;

        i = tp_hash(tpm, &skel);

        TAILQ_FOREACH(elem, &tpm->tpm_elems[i], link) {
                int r = tp_cmp(&skel, elem);

                if (r == 0) /* Match */
                        return elem;
                else if (r > 0)
                        break; /* Not found, break out and add it
                                * at this position. */
        }

        /* Create new tp and add it to its sorted location in the bucket. */
        tp = calloc(1, sizeof(*tp));
        tp->partition = partition;
        tp->topic = ic_strdup(topic);

        timebuckets_init(&tp->tbs, 31, win_size);

        if (elem)
                TAILQ_INSERT_BEFORE(elem, tp, link);
        else
                TAILQ_INSERT_TAIL(&tpm->tpm_elems[i], tp, link);

        TAILQ_INSERT_TAIL(&tpm->tpm_flat, tp, flat_link);

        return tp;
}


void tpmap_clear (tpmap_t *tpm) {
        int i;

        TAILQ_INIT(&tpm->tpm_flat);

        for (i = 0 ; i < tpm->tpm_size ; i++) {
                tp_t *tp, *tmp;

                TAILQ_FOREACH_SAFE(tp, &tpm->tpm_elems[i], link, tmp) {
                        tp_destroy(tp);
                }

                TAILQ_INIT(&tpm->tpm_elems[i]);
        }

        TAILQ_INIT(&tpm->tpm_flat);
}

void tpmap_destroy (tpmap_t *tpm) {
        tpmap_clear(tpm);
        free(tpm->tpm_elems);
}

void tpmap_init (tpmap_t *tpm, int size) {
        int i;

        memset(tpm, 0, sizeof(*tpm));
        tpm->tpm_size = size;

        tpm->tpm_elems = malloc(sizeof(*tpm->tpm_elems) * tpm->tpm_size);

        for (i = 0 ; i < tpm->tpm_size ; i++)
                TAILQ_INIT(&tpm->tpm_elems[i]);

        TAILQ_INIT(&tpm->tpm_flat);
}
